/* See LICENSE file for copyright and license details. */
#include "bwm.h"
#include <stdio.h>
#include <stdlib.h>
#include <X11/keysym.h>
#include <X11/Xatom.h>

/* static */

typedef struct {
	unsigned long mod;
	KeySym keysym;
	void (*func)(const char *arg);
	const char *arg;
} Key;

KEYS

#define CLEANMASK(mask) (mask & ~(numlockmask | LockMask))
#define MOUSEMASK		(BUTTONMASK | PointerMotionMask)

static Client *
getclient(Window w) {
	Client *c;

	for(c = clients; c && c->win != w; c = c->next);
	return c;
}

static void
layersmouse(void) {
	Client *c, *psel;
	int x1, y1, di, dir[DirLast];
	int *drag = NULL;
	unsigned int dui, layerosp = 0;
	Window dummy;
	XEvent ev;
	XButtonPressedEvent *bev = &ev.xbutton;

	psel = sel;
	if(XGrabPointer(dpy, root, False, MOUSEMASK, GrabModeAsync, GrabModeAsync,
			None, cursor[CurMove], CurrentTime) != GrabSuccess)
		return;
	XQueryPointer(dpy, root, &dummy, &dummy, &x1, &y1, &di, &di, &dui);
	for(;;) {
		XMaskEvent(dpy, MOUSEMASK | ExposureMask | SubstructureRedirectMask, &ev);
		switch (ev.type) {
		case ButtonPress:
			if(bev->button == Button1)
				if(layeros > 0) {
					for(c = clients; c; c = c->next)
						if(c->layer == layeros + 1)
							c->layer = 1;
						else if(c->layer > layeros + 1)
							c->layer--;
					lastlayer--;
				}
				else
				for(c = clients; c; c = c->next)
					if(isvisible(c) && c != sel) {
						for(c = clients; c; c = c->next)
							if(c != sel) {
								c->layer++;
								lastlayer = (c->layer > lastlayer) ? c->layer : lastlayer;
							}
						break;
					}
			layeros = 0;
			tile();
			restack();
			XUngrabPointer(dpy, CurrentTime);
			return;
		case ButtonRelease:
			if(layeros != 0)
				for(c = clients; c; c = c->next) {
					if(c->layer == layeros + 1)
						c->layer = 1;
					else if(c->layer < layeros + 1) {
						c->layer++;
						lastlayer = (c->layer > lastlayer) ? c->layer : lastlayer;
					}
				}
			layeros = 0;
			tile();
			XUngrabPointer(dpy, CurrentTime);
			return;
		case ConfigureRequest:
		case Expose:
		case MapRequest:
			handler[ev.type](&ev);
			break;
		case MotionNotify:
			XSync(dpy, False);
			dir[DirVert] = (abs(ev.xmotion.y - y1) / DRAG);
			dir[DirHoriz] = (abs(ev.xmotion.x - x1) / DRAG);
			if(drag == NULL)
				if((abs(ev.xmotion.x - x1) / DRAG) > (abs(ev.xmotion.y - y1) / DRAG))
					drag = &dir[DirHoriz];
				else if((abs(ev.xmotion.x - x1) / DRAG) < (abs(ev.xmotion.y - y1) / DRAG))
					drag = &dir[DirVert];
			layeros = (drag == NULL) ? 0 : *drag;
 			if(layeros > lastlayer - 1)
				layeros = lastlayer - 1;
			if(layeros != layerosp) {
				tile();
				layerosp = layeros;
			}
			break;
		}
	}
}

static void
movemouse(Client *c) {
	int x1, y1, ocx, ocy, ocw, och, di, nx, ny;
	unsigned int dui;
	Window dummy;
	XEvent ev;

	ocx = nx = c->x;
	ocy = ny = c->y;
	ocw = c->w;
	och = c->h;
	if(XGrabPointer(dpy, root, False, MOUSEMASK, GrabModeAsync, GrabModeAsync,
			None, cursor[CurMove], CurrentTime) != GrabSuccess)
		return;
	XQueryPointer(dpy, root, &dummy, &dummy, &x1, &y1, &di, &di, &dui);
	for(;;) {
		XMaskEvent(dpy, MOUSEMASK | ExposureMask | SubstructureRedirectMask, &ev);
		switch (ev.type) {
		case ButtonPress:
			if(ev.xbutton.button == Button2)
				if(abs(ev.xbutton.x) < 2 && abs(ev.xbutton.y) < 2)
					spawn(CMD1);
				else if(abs(ev.xbutton.x - sw) < 2 && abs(ev.xbutton.y) < 2)
					spawn(CMD2);
				else if(abs(ev.xbutton.x) < 2 && abs(ev.xbutton.y - sh) < 2)
					spawn(CMD3);
				else if(abs(ev.xbutton.x - sw) < 2 && abs(ev.xbutton.y - sh) < 2)
					spawn(CMD4);
				else if(ev.xbutton.button == Button2 && c->isfloating && \
				 (c->y == 0 || c->y == sh - c->h - 2 * c->border || \
				 c->x == 0 || c->x == sw - c->w - 2 * c->border)) {
					sel->isnon = !sel->isnon;
					tile();
				}
			XUngrabPointer(dpy, CurrentTime);
			return;
		case ButtonRelease:
			if(!sel->isfloating)
				zoom(NULL);
			if(sel->isnon) {
				if(nx > 0 && ny > 0 && nx != sw - c->w - 2 * c->border \
				 && ny != sh - c->h - 2 * c->border)
					sel->isnon = False;
				tile();
			}
			XUngrabPointer(dpy, CurrentTime);
			return;
		case ConfigureRequest:
		case Expose:
		case MapRequest:
			handler[ev.type](&ev);
			break;
		case MotionNotify:
			XSync(dpy, False);
			nx = ocx + (ev.xmotion.x - x1);
			ny = ocy + (ev.xmotion.y - y1);
			if(abs(wax + nx) < SNAP)
				nx = wax;
			else if(abs((wax + waw) - (nx + c->w + 2 * c->border)) < SNAP)
				nx = wax + waw - c->w - 2 * c->border;
			if(abs(way - ny) < SNAP)
				ny = way;
			else if(abs((way + wah) - (ny + c->h + 2 * c->border)) < SNAP)
				ny = way + wah - c->h - 2 * c->border;
			if(abs(nx - ocx) > SNAP || abs(ny - ocy) > SNAP || sel->isfloating) {
				if(!sel->isfloating)
					togglefloating(NULL);
				resize(c, (nx > 0) ? nx : 0, (ny > 0) ? ny : 0, \
				(nx > 0) ? ((nx + ocw > sw) ? sw - nx - 2 * c->border : ocw) : ocw + nx, \
				(ny > 0) ? ((ny + och > sh) ? sh - ny - 2 * c->border : och) : och + ny, False);
			}
			break;
		}
	}
}

static void
movenon(Client *c) {
	int ocx, ocy, ocw, och, nx, ny, nw, nh;
	XEvent ev;

	ocx = nx = c->x;
	ocy = ny = c->y;
	ocw = c->w;
	och = c->h;
	if(XGrabPointer(dpy, root, False, MOUSEMASK, GrabModeAsync, GrabModeAsync,
			None, cursor[CurMove], CurrentTime) != GrabSuccess)
		return;
	for(;;) {
		XMaskEvent(dpy, MOUSEMASK | ExposureMask | SubstructureRedirectMask, &ev);
		switch (ev.type) {
		case ButtonRelease:
			if(!sel->isnon)
				if(sel->isfloating) {
					togglefloating(NULL);
					zoom(NULL);
				}
				else if(sel->x == wax)
					incnmaster(NULL);
				else
					incmasterw(NULL);
			else
				tile();
			XUngrabPointer(dpy, CurrentTime);
			return;
		case ConfigureRequest:
		case Expose:
		case MapRequest:
			handler[ev.type](&ev);
			break;
		case MotionNotify:
			XSync(dpy, False);
			nx = ev.xmotion.x;
			ny = ev.xmotion.y;
			nw = c->w;
			nh = c->h;

			if(!sel->isnon)
				sel->isnon = True;
			if(!sel->isfloating)
					togglefloating(NULL);

			if(nx < ny && nx < sw - nx && nx < sh - ny) {
				nw = nx - 2 * c->border;
				nx = 0;
				ny = ocy;
				nh = och;
			}
			else if(sw - nx < ny && sw - nx < nx && sw - nx < sh - ny) {
				nw = sw - nx - 2 * c->border;
				ny = ocy;
				nh = och;
			}
			else if(sh - ny < ny && sh - ny < nx && sh - ny < sw - nx) {
				nh = sh - ny - 2 * c->border;
				nx = ocx;
				nw = ocw;
			}
			else if(ny < nx && ny < sw - nx && ny < sh - ny) {
				nh = ny - 2 * c->border;
				ny = 0;
				nx = ocx;
				nw = ocw;
			}
			resize(c, nx, ny, nw, nh, False);
			break;
		}
	}
}

static void
resizemouse(Client *c) {
	int ocx, ocy, ocw, och;
	int nw, nh;
	XEvent ev;

	ocx = c->x;
	ocy = c->y;
	ocw = c->w;
	och = c->h;
	if(XGrabPointer(dpy, root, False, MOUSEMASK, GrabModeAsync, GrabModeAsync,
			None, cursor[CurResize], CurrentTime) != GrabSuccess)
		return;
	XWarpPointer(dpy, None, c->win, 0, 0, 0, 0, c->w + c->border - 1, c->h + c->border - 1);
	for(;;) {
		XMaskEvent(dpy, MOUSEMASK | ExposureMask | SubstructureRedirectMask , &ev);
		switch(ev.type) {
		case ButtonRelease:
			XWarpPointer(dpy, None, c->win, 0, 0, 0, 0,
					c->w + c->border - 1, c->h + c->border - 1);
			if(c->isnon) {
				if(c->x != 0 && c->y != 0)
					resize(c, (c->x + ocw - c->w > 0) ? c->x + ocw - c->w : 0 , \
					 (c->y + och - c->h > 0) ? c->y + och - c->h : 0, c->w, c->h, True);
				tile();
			}
			XUngrabPointer(dpy, CurrentTime);
			while(XCheckMaskEvent(dpy, EnterWindowMask, &ev));
			return;
		case ConfigureRequest:
		case Expose:
		case MapRequest:
			handler[ev.type](&ev);
			break;
		case MotionNotify:
			XSync(dpy, False);
			if((nw = ev.xmotion.x - ocx - 2 * c->border + 1) <= 0)
				nw = 1;
			if((nh = ev.xmotion.y - ocy - 2 * c->border + 1) <= 0)
				nh = 1;
			resize(c, c->x, c->y, nw, nh, True);
			break;
		}
	}
}

static void
buttonpress(XEvent *e) {
	static char buf[32];
	unsigned int i, x;
	XButtonPressedEvent *ev = &e->xbutton;

	buf[0] = 0;
	if(!sel && ev->window == root) {
		if(ev->x < sw / 2 && ev->y < sh / 2)
			spawn(CMD1);
		else if(sw - ev->x < sw / 2 && ev->y < sh / 2)
			spawn(CMD2);
		else if(ev->x < sw / 2 && sh - ev->y < sh / 2)
			spawn(CMD3);
		else if(sw - ev->x < sw / 2 && sh - ev->y < sh / 2)
			spawn(CMD4);
	}
	else if(sel) {
		CLEANMASK(ev->state);
		if(ev->x > sel->x + sel->border && ev->x < sel->x + sel->w - sel->border \
			&& ev->y > sel->y + sel->border && ev->y < sel->y + sel->h - sel->border)
			return;
		if(ev->button == Button1) {
			if(ev->x < sel->x + sel->border || ev->y < sel->y + sel->border) {
				restack();
				movemouse(sel);
			}
			else if(!sel->isfloating) {
				restack();
				movemouse(sel);
			}
			else  if(!sel->isfixed) {
				restack();
				resizemouse(sel);
			}
		}
		else if(ev->button == Button2) {
			restack();
/*			if(sel->isfloating) {
				togglefloating(NULL);
				zoom(NULL);
			}
			else if(sel->x == wax)
				incnmaster(NULL);
			else
				incmasterw(NULL); */
			movenon(sel);
		} 
		else if(ev->button == Button3) {
			restack();
			layersmouse();
		}
	}
}

static void
configurerequest(XEvent *e) {
	Client *c;
	XConfigureRequestEvent *ev = &e->xconfigurerequest;
	XWindowChanges wc;

	if((c = getclient(ev->window))) {
		if(ev->value_mask & CWBorderWidth)
			c->border = ev->border_width;
		if(c->isfixed || c->isfloating) {
			if(ev->value_mask & CWX)
				c->x = ev->x;
			if(ev->value_mask & CWY)
				c->y = ev->y;
			if(ev->value_mask & CWWidth)
				c->w = ev->width;
			if(ev->value_mask & CWHeight)
				c->h = ev->height;
			if((c->x + c->w) > sw && c->isfloating)
				c->x = sw / 2 - c->w / 2; /* center in x direction */
			if((c->y + c->h) > sh && c->isfloating)
				c->y = sh / 2 - c->h / 2; /* center in y direction */
			if((ev->value_mask & (CWX | CWY))
			&& !(ev->value_mask & (CWWidth | CWHeight)))
				configure(c);
			if(isvisible(c))
				XMoveResizeWindow(dpy, c->win, c->x, c->y, c->w, c->h);
		}
		else
			configure(c);
	}
	else {
		wc.x = ev->x;
		wc.y = ev->y;
		wc.width = ev->width;
		wc.height = ev->height;
		wc.border_width = ev->border_width;
		wc.sibling = ev->above;
		wc.stack_mode = ev->detail;
		XConfigureWindow(dpy, ev->window, ev->value_mask, &wc);
	}
	XSync(dpy, False);
}

static void
configurenotify(XEvent *e) {
	XConfigureEvent *ev = &e->xconfigure;

	if (ev->window == root && (ev->width != sw || ev->height != sh)) {
		sw = ev->width;
		sh = ev->height;
		tile();
	}
}

static void
destroynotify(XEvent *e) {
	Client *c;
	XDestroyWindowEvent *ev = &e->xdestroywindow;

	if((c = getclient(ev->window)))
		unmanage(c);
}

static void
enternotify(XEvent *e) {
	Client *c;
	XCrossingEvent *ev = &e->xcrossing;

	if(ev->mode != NotifyNormal || ev->detail == NotifyInferior)
		return;
	if((c = getclient(ev->window)))
		focus(c);
	else if(ev->window == root) {
		selscreen = True;
		focus(NULL);
	}
}

static void
keypress(XEvent *e) {
	static unsigned int len = sizeof key / sizeof key[0];
	unsigned int i;
	KeySym keysym;
	XKeyEvent *ev = &e->xkey;

	keysym = XKeycodeToKeysym(dpy, (KeyCode)ev->keycode, 0);
	for(i = 0; i < len; i++)
		if(keysym == key[i].keysym
		&& CLEANMASK(key[i].mod) == CLEANMASK(ev->state))
		{
			if(key[i].func)
				key[i].func(key[i].arg);
		}
}

static void
leavenotify(XEvent *e) {
	XCrossingEvent *ev = &e->xcrossing;

	if((ev->window == root) && !ev->same_screen) {
		selscreen = False;
		focus(NULL);
	}
}

static void
mappingnotify(XEvent *e) {
	XMappingEvent *ev = &e->xmapping;

	XRefreshKeyboardMapping(ev);
	if(ev->request == MappingKeyboard)
		grabkeys();
}

static void
maprequest(XEvent *e) {
	static XWindowAttributes wa;
	XMapRequestEvent *ev = &e->xmaprequest;

	if(!XGetWindowAttributes(dpy, ev->window, &wa))
		return;
	if(wa.override_redirect)
		return;
	if(!getclient(ev->window))
		manage(ev->window, &wa);
}

static void
propertynotify(XEvent *e) {
	Client *c;
	Window trans;
	XPropertyEvent *ev = &e->xproperty;

	if(ev->state == PropertyDelete)
		return; /* ignore */
	if((c = getclient(ev->window))) {
		switch (ev->atom) {
			default: break;
			case XA_WM_TRANSIENT_FOR:
				XGetTransientForHint(dpy, c->win, &trans);
				if(!c->isfloating && (c->isfloating = (getclient(trans) != NULL)))
					tile();
				break;
			case XA_WM_NORMAL_HINTS:
				updatesizehints(c);
				break;
		}
	}
}

static void
unmapnotify(XEvent *e) {
	Client *c;
	XUnmapEvent *ev = &e->xunmap;

	if((c = getclient(ev->window)))
		unmanage(c);
}

/* extern */

void (*handler[LASTEvent]) (XEvent *) = {
	[ButtonPress] = buttonpress,
	[ConfigureRequest] = configurerequest,
	[ConfigureNotify] = configurenotify,
	[DestroyNotify] = destroynotify,
	[EnterNotify] = enternotify,
	[LeaveNotify] = leavenotify,
	[KeyPress] = keypress,
	[MappingNotify] = mappingnotify,
	[MapRequest] = maprequest,
	[PropertyNotify] = propertynotify,
	[UnmapNotify] = unmapnotify
};

void
grabkeys(void) {
	static unsigned int len = sizeof key / sizeof key[0];
	unsigned int i;
	KeyCode code;

	XUngrabKey(dpy, AnyKey, AnyModifier, root);
	for(i = 0; i < len; i++) {
		code = XKeysymToKeycode(dpy, key[i].keysym);
		XGrabKey(dpy, code, key[i].mod, root, True,
				GrabModeAsync, GrabModeAsync);
		XGrabKey(dpy, code, key[i].mod | LockMask, root, True,
				GrabModeAsync, GrabModeAsync);
		XGrabKey(dpy, code, key[i].mod | numlockmask, root, True,
				GrabModeAsync, GrabModeAsync);
		XGrabKey(dpy, code, key[i].mod | numlockmask | LockMask, root, True,
				GrabModeAsync, GrabModeAsync);
	}
}

