/* See LICENSE file for copyright and license details. */
#include "bwm.h"
#include <stdlib.h>

/* static */

static unsigned int masterw = MASTERWIDTH;
static unsigned int nmaster = NMASTER;

/* extern */

void
focusclient(const char *arg) {
	Client *c;
   
	if(!sel || !arg)
		return;
	if(atoi(arg) < 0) {
		for(c = sel->prev; c && !isvisible(c); c = c->prev);
		if(!c) {
			for(c = clients; c && c->next; c = c->next);
			for(; c && !isvisible(c); c = c->prev);
		}
	}
	else {
		for(c = sel->next; c && !isvisible(c); c = c->next);
		if(!c)
			for(c = clients; c && !isvisible(c); c = c->next);
	}
	if(c) {
		focus(c);
		restack();
	}
}

void
incmasterw(const char *arg) {
	masterw = (masterw == 664) ? 536 : masterw + 64;
	tile();
}

void
incnmaster(const char *arg) {
	nmaster = (nmaster == 3) ? 1 : nmaster + 1;
	if(sel)
		tile();
}

Client *
nexttiled(Client *c) {
	for(; c && (c->isfloating || !isvisible(c)); c = c->next);
	return c;
}

void
restack(void) {
	Client *c;
	XEvent ev;
	XWindowChanges wc;

	if(!sel)
		return;
	for(c = clients; c; c = c->next)
		if(c->isfloating && isvisible(c))
			XRaiseWindow(dpy, c->win);
	if(sel->isfloating)
		XRaiseWindow(dpy, sel->win);
	wc.stack_mode = Below;
	if(!sel->isfloating) {
		XConfigureWindow(dpy, sel->win, CWSibling | CWStackMode, &wc);
		wc.sibling = sel->win;
	}
	for(c = nexttiled(clients); c; c = nexttiled(c->next)) {
		if(c == sel)
			continue;
		XConfigureWindow(dpy, c->win, CWSibling | CWStackMode, &wc);
		wc.sibling = c->win;
	}
	XSync(dpy, False);
	while(XCheckMaskEvent(dpy, EnterWindowMask, &ev));
}

void
tile(void) {
	unsigned int i, n, nx, ny, nw, nh, mw, mh, tw, th, nonw, nonh;
	Client *c;

	/* non overlaping windows stuff */
	wax = sx;
	way = sy;
	nonh = 0;
	nonw = 0;
	for(c = clients; c; c = c->next)
		if(c->isnon && c->isfloating && isvisible(c))
			if(c->y == 0 && c->h < sh - 2 * c->border) {
				way = (c->h + 2 * c->border > way) ? c->h + 2 * c->border : way;
			}
			else if(c->y == sh - c->h - 2 * c->border && c->h < sh - 2 * c->border) {
				nonh = (c->h + 2 * c->border > nonh) ? c->h + 2 * c->border : nonh;
			}
			else if(c->x == 0) {
				wax = (c->w + 2 * c->border > wax) ? c->w + 2 * c->border : wax;
			}
			else if(c->x == sw - c->w - 2 * c->border) {
				nonw = (c->w + 2 * c->border > nonw) ? c->w + 2 * c->border : nonw;
			}
	waw = sw - wax - nonw;
	wah = sh - way - nonh;

	for(n = 0, c = nexttiled(clients); c; c = nexttiled(c->next))
		n++;
	/* window geoms */
	mh = (n > nmaster) ? wah / nmaster : wah / (n > 0 ? n : 1);
	mw = (n > nmaster) ? (waw * masterw) / 1000 : waw;
	th = (n > nmaster) ? wah / (n - nmaster) : 0;
	tw = waw - mw;

	for(i = 0, c = clients; c; c = c->next)
		if(isvisible(c)) {
			unban(c);
			if(c->isfloating)
				continue;
			nx = wax;
			ny = way;
			if(i < nmaster) {
				ny += i * mh;
				nw = mw - 2 * c->border;
				nh = mh;
				if(i + 1 == (n < nmaster ? n : nmaster)) /* remainder */
					nh = wah - mh * i;
				nh -= 2 * c->border;
			}
			else {  /* tile window */
				nx += mw;
				nw = tw - 2 * c->border;
				if(th > 2 * c->border) {
					ny += (i - nmaster) * th;
					nh = th;
					if(i + 1 == n) /* remainder */
						nh = wah - th * (i - nmaster);
					nh -= 2 * c->border;
				}
				else /* fallback if th <= 2 * c->border */
					nh = wah - 2 * c->border;
			}
			resize(c, nx, ny, nw, nh, False);
			i++;
		}
		else
			ban(c);
	focus(NULL);
	restack();
}

void
zoom(const char *arg) {
	Client *c;

	if(!sel || sel->isfloating)
		return;
	if((c = sel) == nexttiled(clients))
		if(!(c = nexttiled(c->next)))
			return;
	detach(c);
	attach(c);
	focus(c);
	tile();
}
